/*
Copyright 2018 The Knative Authors
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package queue

import "fmt"

type token struct{}

// Breaker is a component that enforces a concurrency limit on the
// execution of a function. It also maintains a queue of function
// executions in excess of the concurrency limit. Function call attempts
// beyond the limit of the queue are failed immediately.
type Breaker struct {
	pendingRequests chan token
	activeRequests  chan token
}

// NewBreaker creates a Breaker with the desired queue depth and
// concurrency limit.
func NewBreaker(queueDepth, maxConcurrency int32) *Breaker {
	if queueDepth <= 0 {
		panic(fmt.Sprintf("Queue depth must be greater than 0. Got %v.", queueDepth))
	}
	if maxConcurrency <= 0 {
		panic(fmt.Sprintf("Max concurrency must be greater than 0. Got %v.", maxConcurrency))
	}
	return &Breaker{
		pendingRequests: make(chan token, queueDepth),
		activeRequests:  make(chan token, maxConcurrency),
	}
}

// Maybe conditionally executes thunk based on the Breaker concurrency
// and queue parameters. If the concurrency limit and queue capacity are
// already consumed, Maybe returns immediately without calling thunk. If
// the thunk was executed, Maybe returns true, else false.
func (b *Breaker) Maybe(thunk func()) bool {
	var t token
	select {
	default:
		// Pending request queue is full.  Report failure.
		return false
	case b.pendingRequests <- t:
		// Pending request has capacity.
		// Wait for capacity in the active queue.
		b.activeRequests <- t
		// Release capacity in the pending request queue.
		<-b.pendingRequests
		// Defer releasing capacity in the active request queue.
		defer func() { <-b.activeRequests }()
		// Do the thing.
		thunk()
		// Report success
		return true
	}
}
